from datetime import date, datetime, time
from pathlib import Path
from typing import TYPE_CHECKING, cast

from adaptix import loader
from adaptix.provider import Provider

from dature.expansion.env_expand import expand_env_vars
from dature.protocols import ValidatorProtocol
from dature.sources_loader.base import BaseLoader
from dature.sources_loader.loaders import (
    bool_loader,
    bytearray_from_json_string,
    date_from_string,
    datetime_from_string,
    float_from_string,
    none_from_empty_string,
    optional_from_empty_string,
    str_from_scalar,
    time_from_string,
)
from dature.types import (
    DotSeparatedPath,
    ExpandEnvVarsMode,
    FieldMapping,
    FieldValidators,
    FileOrStream,
    JSONValue,
    NameStyle,
)

if TYPE_CHECKING:
    from dature.metadata import TypeLoader


def _set_nested(d: dict[str, JSONValue], keys: list[str], value: str) -> None:
    for key in keys[:-1]:
        d = cast("dict[str, JSONValue]", d.setdefault(key, {}))
    d[keys[-1]] = value


class DockerSecretsLoader(BaseLoader):
    display_name = "docker_secrets"

    def __init__(  # noqa: PLR0913
        self,
        *,
        prefix: DotSeparatedPath | None = None,
        split_symbols: str = "__",
        name_style: NameStyle | None = None,
        field_mapping: FieldMapping | None = None,
        root_validators: tuple[ValidatorProtocol, ...] | None = None,
        validators: FieldValidators | None = None,
        expand_env_vars: ExpandEnvVarsMode = "default",
        type_loaders: "tuple[TypeLoader, ...]" = (),
    ) -> None:
        self._split_symbols = split_symbols
        super().__init__(
            prefix=prefix,
            name_style=name_style,
            field_mapping=field_mapping,
            root_validators=root_validators,
            validators=validators,
            expand_env_vars=expand_env_vars,
            type_loaders=type_loaders,
        )

    def _additional_loaders(self) -> list[Provider]:
        return [
            loader(str, str_from_scalar),
            loader(float, float_from_string),
            loader(date, date_from_string),
            loader(datetime, datetime_from_string),
            loader(time, time_from_string),
            loader(bytearray, bytearray_from_json_string),
            loader(type(None), none_from_empty_string),
            loader(str | None, optional_from_empty_string),
            loader(bool, bool_loader),
        ]

    def _load(self, path: FileOrStream) -> JSONValue:
        if not isinstance(path, Path):
            msg = "DockerSecretsLoader does not support file-like objects"
            raise TypeError(msg)

        result: dict[str, JSONValue] = {}

        for entry in sorted(path.iterdir()):
            if not entry.is_file():
                continue

            key = entry.name.lower()
            value = entry.read_text().strip()

            if self._prefix and not key.startswith(self._prefix.lower()):
                continue

            if self._prefix:
                key = key[len(self._prefix) :]

            parts = key.split(self._split_symbols)
            if len(parts) > 1:
                _set_nested(result, parts, value)
            else:
                result[key] = value

        return result

    def _pre_processing(self, data: JSONValue) -> JSONValue:
        expanded = expand_env_vars(data, mode=self._expand_env_vars_mode)
        return self._parse_string_values(expanded)



import io
from datetime import date, datetime, time
from typing import TextIO, cast

import json5
from adaptix import loader
from adaptix.provider import Provider

from dature.path_finders.json5_ import Json5PathFinder
from dature.sources_loader.base import BaseLoader
from dature.sources_loader.loaders import (
    bytearray_from_string,
    date_from_string,
    datetime_from_string,
    float_from_string,
    time_from_string,
)
from dature.sources_loader.loaders.json5_ import str_from_json_identifier
from dature.types import BINARY_IO_TYPES, TEXT_IO_TYPES, FileOrStream, JSONValue


class Json5Loader(BaseLoader):
    display_name = "json5"
    path_finder_class = Json5PathFinder

    def _additional_loaders(self) -> list[Provider]:
        return [
            loader(str, str_from_json_identifier),
            loader(float, float_from_string),
            loader(date, date_from_string),
            loader(datetime, datetime_from_string),
            loader(time, time_from_string),
            loader(bytearray, bytearray_from_string),
        ]

    def _load(self, path: FileOrStream) -> JSONValue:
        if isinstance(path, TEXT_IO_TYPES):
            return cast("JSONValue", json5.load(cast("TextIO", path)))
        if isinstance(path, BINARY_IO_TYPES):
            return cast("JSONValue", json5.load(io.TextIOWrapper(cast("io.BufferedReader", path))))
        with path.open() as file_:
            return cast("JSONValue", json5.load(file_))



from contextlib import suppress
from dataclasses import dataclass
from pathlib import Path

from dature.errors.exceptions import LineRange, SourceLocation
from dature.masking.masking import mask_env_line
from dature.path_finders.base import PathFinder


@dataclass(frozen=True)
class ErrorContext:
    dataclass_name: str
    loader_type: str
    file_path: Path | None
    prefix: str | None
    split_symbols: str
    path_finder_class: type[PathFinder] | None
    secret_paths: frozenset[str] = frozenset()
    mask_secrets: bool = False


def read_file_content(file_path: Path | None) -> str | None:
    if file_path is None:
        return None

    with suppress(OSError):
        return file_path.read_text()

    return None


def _build_env_var_name(
    field_path: list[str],
    prefix: str | None,
    split_symbols: str,
) -> str:
    var_name = split_symbols.join(part.upper() for part in field_path)
    if prefix is not None:
        return prefix + var_name
    return var_name


def _build_search_path(field_path: list[str], prefix: str | None) -> list[str]:
    if not prefix:
        return field_path
    prefix_parts = prefix.split(".")
    return prefix_parts + field_path


def _find_env_line(content: str, var_name: str) -> SourceLocation:
    for i, line in enumerate(content.splitlines(), 1):
        stripped = line.strip()
        if not stripped or stripped.startswith("#"):
            continue
        if "=" not in stripped:
            continue
        key = stripped.split("=", 1)[0].strip()
        if key == var_name:
            return SourceLocation(
                source_type="envfile",
                file_path=None,
                line_range=LineRange(start=i, end=i),
                line_content=[stripped],
                env_var_name=var_name,
            )
    return SourceLocation(
        source_type="envfile",
        file_path=None,
        line_range=None,
        line_content=None,
        env_var_name=var_name,
    )


def _empty_file_location(loader_type: str, file_path: Path | None) -> SourceLocation:
    return SourceLocation(
        source_type=loader_type,
        file_path=file_path,
        line_range=None,
        line_content=None,
        env_var_name=None,
    )


def _strip_common_indent(raw_lines: list[str]) -> list[str]:
    indents = [len(line) - len(line.lstrip()) for line in raw_lines if line.strip()]
    if not indents:
        return raw_lines
    min_indent = min(indents)
    return [line[min_indent:] for line in raw_lines]


def _resolve_file_location(
    field_path: list[str],
    loader_type: str,
    file_path: Path | None,
    file_content: str | None,
    prefix: str | None,
    path_finder_class: type[PathFinder] | None,
) -> SourceLocation:
    if file_content is None or not field_path:
        return _empty_file_location(loader_type, file_path)

    if path_finder_class is None:
        return _empty_file_location(loader_type, file_path)

    search_path = _build_search_path(field_path, prefix)
    finder = path_finder_class(file_content)
    line_range = finder.find_line_range(search_path)
    if line_range is None:
        return _empty_file_location(loader_type, file_path)

    lines = file_content.splitlines()
    content_lines: list[str] | None = None
    if 0 < line_range.start <= len(lines):
        end = min(line_range.end, len(lines))
        raw = lines[line_range.start - 1 : end]
        content_lines = _strip_common_indent(raw)

    return SourceLocation(
        source_type=loader_type,
        file_path=file_path,
        line_range=line_range,
        line_content=content_lines,
        env_var_name=None,
    )


def _ranges_overlap(a: LineRange, b: LineRange) -> bool:
    return a.start <= b.end and b.start <= a.end


def _secret_overlaps_lines(
    *,
    file_content: str,
    line_range: LineRange,
    secret_paths: frozenset[str],
    prefix: str | None,
    path_finder_class: type[PathFinder],
) -> bool:
    finder = path_finder_class(file_content)
    for secret_path in secret_paths:
        search_path = _build_search_path(secret_path.split("."), prefix)
        secret_range = finder.find_line_range(search_path)
        if secret_range is not None and _ranges_overlap(line_range, secret_range):
            return True
    return False


def resolve_source_location(
    field_path: list[str],
    ctx: ErrorContext,
    file_content: str | None,
) -> SourceLocation:
    is_secret = ".".join(field_path) in ctx.secret_paths

    if ctx.loader_type == "env":
        env_var_name = _build_env_var_name(field_path, ctx.prefix, ctx.split_symbols)
        return SourceLocation(
            source_type="env",
            file_path=None,
            line_range=None,
            line_content=None,
            env_var_name=env_var_name,
        )

    if ctx.loader_type == "envfile":
        env_var_name = _build_env_var_name(field_path, ctx.prefix, ctx.split_symbols)
        if file_content is not None:
            location = _find_env_line(file_content, env_var_name)
            line_content = location.line_content
            if is_secret and line_content is not None:
                line_content = [mask_env_line(line) for line in line_content]
            return SourceLocation(
                source_type="envfile",
                file_path=ctx.file_path,
                line_range=location.line_range,
                line_content=line_content,
                env_var_name=env_var_name,
            )
        return SourceLocation(
            source_type="envfile",
            file_path=ctx.file_path,
            line_range=None,
            line_content=None,
            env_var_name=env_var_name,
        )

    if ctx.loader_type == "docker_secrets":
        secret_name = ctx.split_symbols.join(field_path)
        if ctx.prefix is not None:
            secret_name = ctx.prefix + secret_name
        secret_file = ctx.file_path / secret_name if ctx.file_path is not None else None
        return SourceLocation(
            source_type="docker_secrets",
            file_path=secret_file,
            line_range=None,
            line_content=None,
            env_var_name=None,
        )

    location = _resolve_file_location(
        field_path=field_path,
        loader_type=ctx.loader_type,
        file_path=ctx.file_path,
        file_content=file_content,
        prefix=ctx.prefix,
        path_finder_class=ctx.path_finder_class,
    )
    should_mask = is_secret
    if (
        not should_mask
        and ctx.secret_paths
        and location.line_range is not None
        and ctx.path_finder_class is not None
        and file_content is not None
    ):
        should_mask = _secret_overlaps_lines(
            file_content=file_content,
            line_range=location.line_range,
            secret_paths=ctx.secret_paths,
            prefix=ctx.prefix,
            path_finder_class=ctx.path_finder_class,
        )
    if should_mask and location.line_content is not None:
        masked_lines = [mask_env_line(line) for line in location.line_content]
        return SourceLocation(
            source_type=location.source_type,
            file_path=location.file_path,
            line_range=location.line_range,
            line_content=masked_lines,
            env_var_name=location.env_var_name,
        )
    return location



"""Tests for json5_ module (Json5Loader)."""

from dataclasses import dataclass
from pathlib import Path

from dature import LoadMetadata, load
from dature.sources_loader.json5_ import Json5Loader
from examples.all_types_dataclass import (
    EXPECTED_ALL_TYPES,
    AllPythonTypesCompact,
)
from tests.sources_loader.checker import assert_all_types_equal


class TestJson5Loader:
    """Tests for Json5Loader class."""

    def test_comprehensive_type_conversion(self, all_types_json5_file: Path):
        """Test loading JSON5 with full type coercion to dataclass."""
        result = load(LoadMetadata(file_=all_types_json5_file, loader=Json5Loader), AllPythonTypesCompact)

        assert_all_types_equal(result, EXPECTED_ALL_TYPES)

    def test_json5_with_prefix(self, prefixed_json5_file: Path):
        @dataclass
        class PrefixedConfig:
            name: str
            port: int
            debug: bool
            environment: str

        expected_data = PrefixedConfig(
            name="PrefixedApp",
            port=9000,
            debug=False,
            environment="production",
        )

        result = load(
            LoadMetadata(file_=prefixed_json5_file, loader=Json5Loader, prefix="app"),
            PrefixedConfig,
        )

        assert result == expected_data

    def test_json5_empty_object(self, tmp_path: Path):
        """Test loading empty JSON5 object."""
        json5_file = tmp_path / "empty.json5"
        json5_file.write_text("{}")

        loader = Json5Loader()
        data = loader._load(json5_file)

        assert data == {}

    def test_json5_env_var_substitution(self, tmp_path: Path, monkeypatch):
        monkeypatch.setenv("DB_HOST", "db.example.com")
        monkeypatch.setenv("DB_PORT", "5432")

        json5_file = tmp_path / "env.json5"
        json5_file.write_text('{host: "$DB_HOST", port: "$DB_PORT"}')

        @dataclass
        class DbConfig:
            host: str
            port: int

        result = load(LoadMetadata(file_=json5_file, loader=Json5Loader), DbConfig)

        assert result.host == "db.example.com"
        assert result.port == 5432

    def test_json5_env_var_partial_substitution(self, tmp_path: Path, monkeypatch):
        monkeypatch.setenv("HOST", "localhost")
        monkeypatch.setenv("PORT", "8080")

        json5_file = tmp_path / "env.json5"
        json5_file.write_text('{url: "http://${HOST}:${PORT}/api"}')

        @dataclass
        class Config:
            url: str

        result = load(LoadMetadata(file_=json5_file, loader=Json5Loader), Config)

        assert result.url == "http://localhost:8080/api"

    def test_json5_dollar_sign_mid_string_existing_var(self, tmp_path: Path, monkeypatch):
        monkeypatch.setenv("abc", "replaced")

        json5_file = tmp_path / "dollar.json5"
        json5_file.write_text('{value: "prefix$abc/suffix"}')

        @dataclass
        class Config:
            value: str

        result = load(LoadMetadata(file_=json5_file, loader=Json5Loader), Config)

        assert result.value == "prefixreplaced/suffix"

    def test_json5_dollar_sign_mid_string_missing_var(self, tmp_path: Path, monkeypatch):
        monkeypatch.delenv("nonexistent", raising=False)

        json5_file = tmp_path / "dollar.json5"
        json5_file.write_text('{value: "prefix$nonexistent/suffix"}')

        @dataclass
        class Config:
            value: str

        result = load(LoadMetadata(file_=json5_file, loader=Json5Loader), Config)

        assert result.value == "prefix$nonexistent/suffix"
